/* vi:set ts=8 sts=8 sw=8:
 *
 * PMS  <<Practical Music Search>>
 * Copyright (C) 2006-2009  Kim Tore Jensen
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include "input.h"
#include "config.h"
#include "pms.h"

extern Pms *		pms;


Input::Input()
{
	this->_mode = INPUT_NORMAL;
	pending = PEND_NONE;
	text.clear();
	searchterm.clear();
	cmdhistory.clear();
	searchhistory.clear();

	winclear();
}

Input::~Input()
{
}

/*
 * Store values when an extra window parameter is needed
 */
void			Input::winstore(pms_window * w)
{
	win = w;
	winparam = param;
	winpend = pending;
	debug("winstore: win=%p, winparam=%s, winpend=%d\n", win, winparam.c_str(), winpend);
}

/*
 * Delete window parameters
 */
void			Input::winclear()
{
	win = NULL;
	winparam = "";
	winpend = PEND_NONE;
}

/*
 * Restore window parameters
 */
bool			Input::winpop()
{
	if (win == NULL) return false;
	debug("winpop: setting param=%s, pending=%d\n", param.c_str(), pending);

	param = winparam;
	pending = winpend;

	return true;
}

/*
 * Go to next history item
 */
bool			Input::gonext()
{
	vector<string>::const_iterator	e;

	if (_mode == INPUT_JUMP)
		e = searchhistory.end();
	else if (_mode == INPUT_COMMAND)
		e = cmdhistory.end();
	else
		return false;

	if (historypos == e)
		return false;

	++historypos;

	if (historypos == e)
	{
		text.clear();
		return false;
	}
	text = *historypos;
	return true;
}

/*
 * Go to previous history item
 */
bool			Input::goprev()
{
	vector<string>::const_iterator	e;

	if (_mode == INPUT_JUMP)
		e = searchhistory.begin();
	else if (_mode == INPUT_COMMAND)
		e = cmdhistory.begin();
	else
		return false;

	if (historypos == e)
		return false;

	--historypos;
	text = *historypos;
	return true;
}

/*
 * Set input mode
 */
void			Input::mode(Input_mode m)
{
	debug("Entering input mode %d\n", m);

	switch(m)
	{
		case INPUT_COMMAND:
			text.clear();
			historypos = cmdhistory.end();
			break;

		case INPUT_JUMP:
			text.clear();
			searchterm.clear();
			historypos = searchhistory.end();
			break;

		default:
			break;
	}

	this->_mode = m;
}

Input_mode		Input::mode()
{
	return this->_mode;
}

int			Input::get_keystroke()
{
	ch = getch();
	if (ch != -1)
	{
		return true;
	}

	return false;
}

pms_pending_keys	Input::dispatch()
{
	this->pending = PEND_NONE;
	this->param.clear();

	if (_mode == INPUT_NORMAL)
		return dispatch_normal();
	else if (_mode == INPUT_LIST)
		return dispatch_list();
	else
		return dispatch_text();

	return PEND_NONE;
}

pms_pending_keys	Input::dispatch_text()
{
	switch(ch)
	{
		case KEY_RESIZE:	/* Window was resized */
			pending = PEND_RESIZE;
			break;
		case 10:		/* Return */
		case 343:		/* Enter (keypad) */
			pending = PEND_TEXT_RETURN;
			break;
		case 21:		/* ^U */
			if (text.size() == 0)
			{
				text.clear();
				pending = PEND_TEXT_ESCAPE;
			}
			else
			{
				text.clear();
				pending = PEND_TEXT_UPDATED;
			}
			break;
		case 27:		/* Escape */
			pending = PEND_RETURN_ESCAPE;
			break;
		case 8:			/* ^H -- backspace */
		case 127:		/* ^? -- delete */
		case KEY_BACKSPACE:
			if (text.size() > 0)
			{
				text.erase(--string::iterator(text.end()));
				pending = PEND_TEXT_UPDATED;
			}
			else
			{
				pending = PEND_TEXT_ESCAPE;
			}
			break;
		case KEY_MOUSE:
			pending = PEND_NONE;
			break;
		default:
			if (ch < 32 || ch >= KEY_CODE_YES)
			{
				pending = pms->bindings->act(ch, &param);
				if (pending != PEND_NONE)
				{
					break;
				}
				debug("Key %3d '%c' pressed in text mode but not textual and not bound.\n", ch, ch);
			}
			text += ch;

			pending = PEND_TEXT_UPDATED;
	}

	return pending;
}

pms_pending_keys	Input::dispatch_list()
{
	if (ch == 10 || ch == 343)
	{
		pending = PEND_RETURN;
		return pending;
	}
	else if (ch == 27)
	{
		pending = PEND_RETURN_ESCAPE;
		return pending;
	}

	return dispatch_normal();
}

pms_pending_keys	Input::dispatch_normal()
{
	MEVENT		mouseevent;

	if (ch == -1)
		return PEND_NONE;

	if (ch == KEY_RESIZE)
	{
		pending = PEND_RESIZE;
		return pending;
	}

	/* Mouse event */
	if (ch == KEY_MOUSE)
	{
		if (getmouse(&mouseevent) == ERR)
		{
			debug("error with getmouse()\n");
			return PEND_NONE;
		}
		debug("mevent x:%d, y:%d, z:%d\n", mouseevent.x, mouseevent.y, mouseevent.z);
		switch (mouseevent.bstate)
		{
			case BUTTON1_PRESSED:
				debug("button 1 down\n");
				return PEND_NONE;
			case BUTTON1_RELEASED:
				debug("button 1 released\n");
				return PEND_NONE;
			case BUTTON1_CLICKED:
				debug("button 1 clicked\n");
				return PEND_NONE;
			case BUTTON1_DOUBLE_CLICKED:
				debug("button 1 doubleclicked\n");
				return PEND_NONE;
			case BUTTON1_TRIPLE_CLICKED:
				debug("button 1 tripleclicked\n");
				return PEND_NONE;
			case BUTTON2_PRESSED:
				debug("button 2 down\n");
				return PEND_NONE;
			case BUTTON2_RELEASED:
				debug("button 2 released\n");
				return PEND_NONE;
			case BUTTON2_CLICKED:
				debug("button 2 clicked\n");
				return PEND_NONE;
			case BUTTON2_DOUBLE_CLICKED:
				debug("button 2 doubleclicked\n");
				return PEND_NONE;
			case BUTTON2_TRIPLE_CLICKED:
				debug("button 2 tripleclicked\n");
				return PEND_NONE;
			case BUTTON3_PRESSED:
				debug("button 3 down\n");
				return PEND_NONE;
			case BUTTON3_RELEASED:
				debug("button 3 released\n");
				return PEND_NONE;
			case BUTTON3_CLICKED:
				debug("button 3 clicked\n");
				return PEND_NONE;
			case BUTTON3_DOUBLE_CLICKED:
				debug("button 3 doubleclicked\n");
				return PEND_NONE;
			case BUTTON3_TRIPLE_CLICKED:
				debug("button 3 tripleclicked\n");
				return PEND_NONE;
			case BUTTON4_PRESSED:
				debug("button 4 down\n");
				return PEND_NONE;
			case BUTTON4_RELEASED:
				debug("button 4 released\n");
				return PEND_NONE;
			case BUTTON4_CLICKED:
				debug("button 4 clicked\n");
				return PEND_NONE;
			case BUTTON4_DOUBLE_CLICKED:
				debug("button 4 doubleclicked\n");
				return PEND_NONE;
			case BUTTON4_TRIPLE_CLICKED:
				debug("button 4 tripleclicked\n");
				return PEND_NONE;
			/*case BUTTON5_PRESSED:
				debug("button 5 down\n");
				return PEND_NONE;
			case BUTTON5_RELEASED:
				debug("button 5 released\n");
				return PEND_NONE;
			case BUTTON5_CLICKED:
				debug("button 5 clicked\n");
				return PEND_NONE;
			case BUTTON5_DOUBLE_CLICKED:
				debug("button 5 doubleclicked\n");
				return PEND_NONE;
			case BUTTON5_TRIPLE_CLICKED:
				debug("button 5 tripleclicked\n");
				return PEND_NONE;*/
			case BUTTON_SHIFT:
				debug("shift\n");
				return PEND_NONE;
			case BUTTON_CTRL:
				debug("control\n");
				return PEND_NONE;
			case BUTTON_ALT:
				debug("alt\n");
				return PEND_NONE;
			case REPORT_MOUSE_POSITION:
				debug("mouse position\n");
				return PEND_NONE;
			case ALL_MOUSE_EVENTS:
				debug("all events\n");
				return PEND_NONE;
			default:
				debug("mevent state unknown: %d\n", mouseevent.bstate);
				return PEND_NONE;
		}
	}

	/* Key pressed */
	pending = pms->bindings->act(ch, &param);

	if (pending == PEND_NONE)
	{
		pms->setstatus(STERR, "Key is not bound.");
		debug("Key %3d '%c' pressed but not bound.\n", ch, ch);
	}

	return pending;
}

/*
 * Saves the current text in history
 */
void		Input::savehistory()
{
	switch(_mode)
	{
		case INPUT_JUMP:
			searchhistory.push_back(text);
			historypos = searchhistory.end();
			break;

		case INPUT_COMMAND:
			cmdhistory.push_back(text);
			historypos = cmdhistory.end();
			break;

		default:
			break;
	}
}

/*
 * Run a command
 */
bool		Input::run(string s, Error & err)
{
	int		pos;

	if (s.size() == 0)
		return true;

	pos = s.find_first_of(" ");
	if (pos > 0)
	{
		param = s.substr(pos + 1);
		s = s.substr(0, pos);
		debug("Running command '%s' with param '%s'\n", s.c_str(), param.c_str());
	}
	else
		debug("Running command '%s' without parameters\n", s.c_str());

	pending = pms->commands->act(s);

	if (pending == PEND_NONE)
	{
		err.code = CERR_UNKNOWN_COMMAND;
		err.str = _("unknown command");
		err.str += " '" + s + "'";
		return false;
	}

	return true;
}
